from typing import Any

from dbus_next import PropertyAccess, Variant
from dbus_next.aio import MessageBus
from dbus_next.service import ServiceInterface, dbus_property, method, signal

from aionowplaying.interface.base import BaseInterface, PropertyName, PlayerProperties, PlaybackProperties, \
    PlaybackPropertyName, LoopStatus, TrackListPropertyName, TrackListProperties


class DBusBeanMapper:
    @staticmethod
    def metadata(metadata: PlaybackProperties.MetadataBean) -> dict:
        metadata_map = dict()
        metadata_map['mpris:trackid'] = Variant('s', metadata.id_)
        metadata_map['mpris:length'] = Variant('x', metadata.duration)
        metadata_map['mpris:artUrl'] = Variant('s', metadata.cover)
        metadata_map['xesam:album'] = Variant('s', metadata.album)
        metadata_map['xesam:albumArtist'] = Variant('as', metadata.albumArtist)
        metadata_map['xesam:artist'] = Variant('as', metadata.artist)
        metadata_map['xesam:asText'] = Variant('s', metadata.lyrics)
        metadata_map['xesam:comment'] = Variant('as', metadata.comments)
        metadata_map['xesam:composer'] = Variant('as', metadata.composer)
        metadata_map['xesam:genre'] = Variant('as', metadata.genre)
        metadata_map['xesam:lyricist'] = Variant('as', metadata.lyricist)
        metadata_map['xesam:title'] = Variant('s', metadata.title)
        metadata_map['xesam:trackNumber'] = Variant('i', metadata.trackNumber)
        metadata_map['xesam:url'] = Variant('s', metadata.url)
        return metadata_map


class MprisPlayerServiceInterface(ServiceInterface):
    def __init__(self, bus_name: str, it: 'Mpris2Interface' = None):
        super().__init__(bus_name)
        self._properties = PlaybackProperties()
        self._it = it

    def set_property(self, name: str, value: Any):
        setattr(self._properties, name, value)
        if isinstance(value, PlaybackProperties.MetadataBean):
            value = DBusBeanMapper.metadata(value)
        self.emit_properties_changed({name: value})

    @dbus_property(access=PropertyAccess.READ, name=PlaybackPropertyName.PlaybackStatus.value)
    def playback_status(self) -> 's':
        return self._properties.PlaybackStatus.value

    @dbus_property(access=PropertyAccess.READWRITE, name=PlaybackPropertyName.LoopStatus.value)
    def loop_status(self) -> 's':
        return self._properties.LoopStatus.value

    @loop_status.setter
    async def loop_status(self, value: 's'):
        if self._properties.CanControl:
            await self._it.on_loop_status(LoopStatus(value))
            self._properties.LoopStatus = LoopStatus(value)

    @dbus_property(access=PropertyAccess.READWRITE, name=PlaybackPropertyName.Rate.value)
    def rate(self) -> 'd':
        return self._properties.Rate

    @rate.setter
    async def rate(self, value: 'd'):
        await self._it.on_rate(value)
        self._properties.Rate = value

    @dbus_property(access=PropertyAccess.READWRITE, name=PlaybackPropertyName.Shuffle.value)
    def shuffle(self) -> 'b':
        return self._properties.Shuffle

    @shuffle.setter
    async def shuffle(self, value: 'b'):
        if self._properties.CanControl:
            await self._it.on_shuffle(value)
            self._properties.Shuffle = value

    @dbus_property(access=PropertyAccess.READ, name=PlaybackPropertyName.Metadata.value)
    def metadata(self) -> 'a{sv}':
        metadata = self._properties.Metadata
        return DBusBeanMapper.metadata(metadata)

    @dbus_property(access=PropertyAccess.READWRITE, name=PlaybackPropertyName.Volume.value)
    def volume(self) -> 'd':
        return self._properties.Volume

    @volume.setter
    async def volume(self, value: 'd'):
        if self._properties.CanControl:
            await self._it.on_volume(value)
            self._properties.Volume = value

    @dbus_property(access=PropertyAccess.READ, name=PlaybackPropertyName.Position.value)
    def position(self) -> 'x':
        return self._properties.Position

    @dbus_property(access=PropertyAccess.READ, name=PlaybackPropertyName.MinimumRate.value)
    def minimum_rate(self) -> 'd':
        return self._properties.MinimumRate

    @dbus_property(access=PropertyAccess.READ, name=PlaybackPropertyName.MaximumRate.value)
    def maximum_rate(self) -> 'd':
        return self._properties.MaximumRate

    @dbus_property(access=PropertyAccess.READ, name=PlaybackPropertyName.CanGoNext.value)
    def can_go_next(self) -> 'b':
        return self._properties.CanGoNext

    @dbus_property(access=PropertyAccess.READ, name=PlaybackPropertyName.CanGoPrevious.value)
    def can_go_previous(self) -> 'b':
        return self._properties.CanGoPrevious

    @dbus_property(access=PropertyAccess.READ, name=PlaybackPropertyName.CanPlay.value)
    def can_play(self) -> 'b':
        return self._properties.CanPlay

    @dbus_property(access=PropertyAccess.READ, name=PlaybackPropertyName.CanPause.value)
    def can_pause(self) -> 'b':
        return self._properties.CanPause

    @dbus_property(access=PropertyAccess.READ, name=PlaybackPropertyName.CanSeek.value)
    def can_seek(self) -> 'b':
        return self._properties.CanSeek

    @dbus_property(access=PropertyAccess.READ, name=PlaybackPropertyName.CanControl.value)
    def can_control(self) -> 'b':
        return self._properties.CanControl

    @signal(name='Seeked')
    async def seeked(self, position: int) -> 'x':
        return position

    @method(name="Next")
    async def next(self):
        if self._properties.CanGoNext:
            await self._it.on_next()

    @method(name="Previous")
    async def previous(self):
        if self._properties.CanGoPrevious:
            await self._it.on_previous()

    @method(name="Pause")
    async def pause(self):
        if self._properties.CanPause:
            await self._it.on_pause()

    @method(name="PlayPause")
    async def play_pause(self):
        if self._properties.CanPause:
            await self._it.on_play_pause()

    @method(name="Stop")
    async def stop(self):
        if self._properties.CanControl:
            await self._it.on_stop()

    @method(name="Play")
    async def play(self):
        if self._properties.CanPlay:
            await self._it.on_play()

    @method(name="Seek")
    async def seek(self, offset: 'x'):
        if self._properties.CanSeek:
            await self._it.on_seek(offset)

    @method(name="OpenUri")
    async def open_uri(self, uri: 's'):
        await self._it.on_open_uri(uri)

    @method(name="SetPosition")
    async def set_position(self, track_id: 'o', position: 'x'):
        if self._properties.CanSeek:
            await self._it.on_set_position(track_id, position)

    def get_property(self, key):
        return getattr(self._properties, key)


class MprisServiceInterface(ServiceInterface):
    def __init__(self, bus_name: str, it: 'Mpris2Interface' = None):
        super().__init__(bus_name)
        self._properties = PlayerProperties()
        self._it = it

    @dbus_property(access=PropertyAccess.READWRITE, name=PropertyName.Fullscreen.value)
    def fullscreen(self) -> 'b':
        return self._properties.Fullscreen

    @fullscreen.setter
    async def fullscreen(self, value: 'b'):
        if self._properties.CanSetFullscreen:
            await self._it.on_fullscreen(value)
            self._properties.Fullscreen = value

    @dbus_property(access=PropertyAccess.READ, name=PropertyName.CanQuit.value)
    def can_quit(self) -> 'b':
        return self._properties.CanQuit

    @dbus_property(access=PropertyAccess.READ, name=PropertyName.CanSetFullscreen.value)
    def can_set_fullscreen(self) -> 'b':
        return self._properties.CanSetFullscreen

    @dbus_property(access=PropertyAccess.READ, name=PropertyName.HasTrackList.value)
    def has_track_list(self) -> 'b':
        return self._properties.HasTrackList

    @dbus_property(access=PropertyAccess.READ, name=PropertyName.CanRaise.value)
    def can_raise(self) -> 'b':
        return self._properties.CanRaise

    @dbus_property(access=PropertyAccess.READ, name=PropertyName.Identity.value)
    def identity(self) -> 's':
        return self._properties.Identity

    @dbus_property(access=PropertyAccess.READ, name=PropertyName.DesktopEntry.value)
    def desktop_entry(self) -> 's':
        return self._properties.DesktopEntry

    @dbus_property(access=PropertyAccess.READ, name=PropertyName.SupportedUriSchemes.value)
    def supported_uri_schemes(self) -> 'as':
        return self._properties.SupportedUriSchemes

    @dbus_property(access=PropertyAccess.READ, name=PropertyName.SupportedMimeTypes.value)
    def supported_mime_types(self) -> 'as':
        return self._properties.SupportedMimeTypes

    @method(name='Raise')
    async def raise_(self):
        if self._properties.CanRaise:
            await self._it.on_raise()

    @method(name='Quit')
    async def quit(self):
        if self._properties.CanQuit:
            await self._it.on_quit()

    def set_property(self, name: str, value: Any):
        setattr(self._properties, name, value)
        self.emit_properties_changed({name: value})

    def get_property(self, key):
        return getattr(self._properties, key)


class MprisTracklistServiceInterface(ServiceInterface):
    def __init__(self, bus_name: str, it: 'Mpris2Interface' = None):
        super().__init__(bus_name)
        self._properties = TrackListProperties()
        self._it = it

    def set_property(self, name: str, value: Any):
        setattr(self._properties, name, value)

    @dbus_property(access=PropertyAccess.READ, name=TrackListPropertyName.CanEditTracks.value)
    def can_edit_tracks(self) -> 'b':
        return self._properties.CanEditTracks

    @dbus_property(access=PropertyAccess.READ, name=TrackListPropertyName.Tracks.value)
    def tracks(self) -> 'ao':
        return self._properties.Tracks

    def get_property(self, key):
        return getattr(self._properties, key)


class Mpris2Interface(BaseInterface):
    def __init__(self, name: str):
        super().__init__(name)
        self._bus_name = f'org.mpris.MediaPlayer2.{name}'
        self._entry_name = 'org.mpris.MediaPlayer2'
        self._player_entry_name = 'org.mpris.MediaPlayer2.Player'
        self._player_tracklist_name = 'org.mpris.MediaPlayer2.TrackList'
        self._object_path = '/org/mpris/MediaPlayer2'
        self._bus = MprisServiceInterface(self._entry_name, it=self)
        self._player_bus = MprisPlayerServiceInterface(self._player_entry_name, it=self)
        self._tracklist_bus = MprisTracklistServiceInterface(self._player_tracklist_name, it=self)

    def set_property(self, name: PropertyName, value: Any):
        self._bus.set_property(name.value, value)

    def set_playback_property(self, name: PlaybackPropertyName, value: Any):
        self._player_bus.set_property(name.value, value)

    def set_tracklist_property(self, name: TrackListPropertyName, value: Any):
        self._tracklist_bus.set_property(name.value, value)

    def get_property(self, name: PropertyName) -> Any:
        return self._bus.get_property(name.value)

    def get_playback_property(self, name: PlaybackPropertyName) -> Any:
        return self._player_bus.get_property(name.value)

    def get_tracklist_property(self, name: TrackListPropertyName) -> Any:
        return self._tracklist_bus.get_property(name.value)

    async def seeked(self, position: int):
        await self._player_bus.seeked(position)

    async def start(self):
        self.dbus = await MessageBus().connect()
        self.dbus.export(self._object_path, self._bus)
        self.dbus.export(self._object_path, self._player_bus)
        await self.dbus.request_name(self._bus_name)
        await self.dbus.wait_for_disconnect()

    async def stop(self):
        self.dbus.disconnect()


if __name__ == '__main__':
    import asyncio

    mp = Mpris2Interface('aionowplaying')
    mp.set_property(PropertyName.CanQuit, True)
    loop = asyncio.get_event_loop()
    loop.run_until_complete(mp.start())



import importlib
import sys
from typing import Type

from aionowplaying.interface.base import BaseInterface

INTERFACES_BY_SYSTEM = {
    'linux': 'aionowplaying.interface.mpris2.Mpris2Interface',
    'win32': 'aionowplaying.interface.windows.WindowsInterface',
    'darwin': 'aionowplaying.interface.macos.MacOSInterface',
}


def select_interface(system: str = None) -> Type[BaseInterface]:
    if system is None:
        system = sys.platform
    name = INTERFACES_BY_SYSTEM.get(system, 'aionowplaying.interface.base.BaseInterface')
    mod = name.rsplit('.', 1)
    return getattr(importlib.import_module(mod[0]), mod[1])



from enum import Enum
from typing import Any, List
from pydantic import BaseModel


class TrackListPropertyName(str, Enum):
    Tracks = 'Tracks'
    CanEditTracks = 'CanEditTracks'


class PropertyName(str, Enum):
    CanQuit = "CanQuit"
    CanSetFullscreen = "CanSetFullscreen"
    CanRaise = "CanRaise"
    HasTrackList = "HasTrackList"
    Identity = "Identity"
    DesktopEntry = "DesktopEntry"
    SupportedUriSchemes = "SupportedUriSchemes"
    SupportedMimeTypes = "SupportedMimeTypes"
    Fullscreen = "Fullscreen"


class PlaybackPropertyName(str, Enum):
    PlaybackStatus = "PlaybackStatus"
    LoopStatus = "LoopStatus"
    Rate = "Rate"
    Shuffle = "Shuffle"
    Metadata = "Metadata"
    Volume = "Volume"
    Position = "Position"
    Duration = "Duration"
    MinimumRate = "MinimumRate"
    MaximumRate = "MaximumRate"
    CanGoNext = "CanGoNext"
    CanGoPrevious = "CanGoPrevious"
    CanPlay = "CanPlay"
    CanPause = "CanPause"
    CanSeek = "CanSeek"
    CanControl = "CanControl"


class PlaybackStatus(str, Enum):
    Playing = "Playing"
    Paused = "Paused"
    Stopped = "Stopped"


class LoopStatus(str, Enum):
    None_ = "None"
    Track = "Track"
    Playlist = "Playlist"


class MediaType(str, Enum):
    Music = "Music"
    Video = "Video"
    Image = "Image"


class TrackListProperties(BaseModel):
    Tracks: List[str] = []
    CanEditTracks: bool = False


class PlayerProperties(BaseModel):
    Fullscreen: bool = False
    CanQuit: bool = False
    CanSetFullscreen: bool = False
    CanRaise: bool = False
    HasTrackList: bool = False
    Identity: str = ""
    DesktopEntry: str = ""
    SupportedUriSchemes: List[str] = []
    SupportedMimeTypes: List[str] = []


class PlaybackProperties(BaseModel):
    class MetadataBean(BaseModel):
        id_: str = ''
        media_type: 'MediaType' = MediaType.Music
        duration: int = 0  # in microseconds
        cover: str = ''
        album: str = ''
        albumArtist: List[str] = []
        artist: List[str] = []
        lyrics: str = ''
        comments: List[str] = []
        composer: List[str] = []
        genre: List[str] = []
        lyricist: List[str] = []
        title: str = "Unknown"
        trackNumber: int = 0
        url: str = ''

    PlaybackStatus: 'PlaybackStatus' = PlaybackStatus.Stopped
    LoopStatus: 'LoopStatus' = LoopStatus.None_
    Rate: float = 1.0
    Shuffle: bool = False
    Metadata: MetadataBean = MetadataBean()
    Volume: float = 1.0
    Position: int = 0  # in microseconds
    Duration: int = 0  # in microseconds
    MinimumRate: float = 1.0
    MaximumRate: float = 1.0
    CanGoNext: bool = False
    CanGoPrevious: bool = False
    CanPlay: bool = False
    CanPause: bool = False
    CanSeek: bool = False
    CanControl: bool = False


class BaseInterface:
    def __init__(self, name: str):
        pass

    async def start(self):
        """
        Call this method to start nowplaying backend.
        Wrap with :meth:`asyncio.ensure_future` if you want to run in background.
        """
        pass

    async def stop(self):
        pass

    async def on_fullscreen(self, fullscreen: bool):
        """
        This will be called when nowplaying backend want to set player fullscreen state.
        This will only be called if you set :attr:`PlayerProperties.CanSetFullscreen` to True.
        :param fullscreen: True if fullscreen, False otherwise.
        :type fullscreen: bool
        """
        pass

    async def on_raise(self):
        """
        This will be called when nowplaying backend want to raise player window.
        This will only be called if you set :attr:`PlayerProperties.CanRaise` to True.
        """
        pass

    async def on_quit(self):
        """
        This will be called when nowplaying backend want to quit player.
        This will only be called if you set :attr:`PlayerProperties.CanQuit` to True.
        """
        pass

    async def on_loop_status(self, status: LoopStatus):
        """
        This will be called when nowplaying backend want to set loop status.
        This will only be called if you set :attr:`PlaybackProperties.CanControl` to True.
        :param status: Loop status.
        :type status: LoopStatus
        """
        pass

    async def on_rate(self, rate: float):
        """
        This will be called when nowplaying backend want to set playback rate.
        The rate is a float value between :attr:`PlaybackProperties.MinimumRate`
        and :attr:`PlaybackProperties.MaximumRate`.
        :param rate: Playback rate.
        :type rate: float
        """
        pass

    async def on_shuffle(self, shuffle: bool):
        """
        This will be called when nowplaying backend want to set shuffle status.
        This will only be called if you set :attr:`PlaybackProperties.CanControl` to True.
        :param shuffle: True if shuffle, False otherwise.
        :type shuffle: bool
        """
        pass

    async def on_volume(self, volume: float):
        """
        This will be called when nowplaying backend want to set playback volume.
        This will only be called if you set :attr:`PlaybackProperties.CanControl` to True.
        :param volume: Volume value between 0.0 and 1.0 (both inclusive).
        :type volume: float
        """
        pass

    async def on_next(self):
        """
        This will be called when nowplaying backend want to play next track.
        This will only be called if you set :attr:`PlaybackProperties.CanGoNext` to True.
        """
        pass

    async def on_previous(self):
        """
        This will be called when nowplaying backend want to play previous track.
        This will only be called if you set :attr:`PlaybackProperties.CanGoPrevious` to True.
        """
        pass

    async def on_pause(self):
        """
        This will be called when nowplaying backend want to pause playback.
        This will only be called if you set :attr:`PlaybackProperties.CanPause` to True.
        """
        pass

    async def on_play_pause(self):
        """
        This will be called when nowplaying backend want to play or pause playback.
        This will only be called if you set :attr:`PlaybackProperties.CanPause` to True.
        """
        if self.get_playback_property(PlaybackPropertyName.PlaybackStatus) == PlaybackStatus.Playing:
            await self.on_pause()
        else:
            await self.on_play()

    async def on_play(self):
        """
        This will be called when nowplaying backend want to play playback.
        This will only be called if you set :attr:`PlaybackProperties.CanPlay` to True.
        """
        pass

    async def on_stop(self):
        pass

    async def on_seek(self, offset: int):
        pass

    async def on_open_uri(self, uri: str):
        pass

    async def on_set_position(self, track_id: str, position: int):
        pass

    async def seeked(self, position: int):
        pass

    def set_property(self, name: PropertyName, value: Any):
        pass

    def set_playback_property(self, name: PlaybackPropertyName, value: Any):
        pass

    def set_tracklist_property(self, name: TrackListPropertyName, value: Any):
        pass

    def get_property(self, name: PropertyName) -> Any:
        pass

    def get_playback_property(self, name: PlaybackPropertyName) -> Any:
        pass

    def get_tracklist_property(self, name: TrackListPropertyName) -> Any:
        pass



__all__ = ['select_interface', 'BaseInterface', 'PropertyName', 'LoopStatus', 'PlaybackPropertyName',
           'PlaybackProperties', 'PlaybackStatus', 'NowPlayingInterface']

from typing import Type

from aionowplaying.interface import select_interface, BaseInterface
from aionowplaying.interface.base import PropertyName, LoopStatus, PlaybackPropertyName, PlaybackProperties, \
    PlaybackStatus

NowPlayingInterface: Type[BaseInterface] = select_interface()
if NowPlayingInterface is None:
    raise NotImplemented()



import asyncio
from typing import Any

from Foundation import NSMutableDictionary
from MediaPlayer import MPRemoteCommandCenter, MPNowPlayingInfoCenter
from MediaPlayer import (
    MPMediaItemPropertyTitle, MPMediaItemPropertyArtist, MPMediaItemPropertyAlbumTitle,
    MPMusicPlaybackStatePlaying, MPMusicPlaybackStatePaused,
    MPMusicPlaybackStateStopped, MPMediaItemPropertyPlaybackDuration,
    MPNowPlayingInfoPropertyPlaybackRate, MPNowPlayingInfoPropertyElapsedPlaybackTime,
    MPRemoteCommandHandlerStatusSuccess, MPNowPlayingInfoPropertyDefaultPlaybackRate,
)

from aionowplaying import (
    BaseInterface, PlaybackPropertyName, PlaybackProperties, PlaybackStatus
)

PlaybackStatusStateMapping = {
    PlaybackStatus.Playing: MPMusicPlaybackStatePlaying,
    PlaybackStatus.Paused: MPMusicPlaybackStatePaused,
    PlaybackStatus.Stopped: MPMusicPlaybackStateStopped,
}


def create_handler(_, handler):
    def handle(_):
        asyncio.create_task(handler())
        return MPRemoteCommandHandlerStatusSuccess
    return handle


class MacOSInterface(BaseInterface):
    def __init__(self, name: str):
        super().__init__(name)
        self.cmd_center = MPRemoteCommandCenter.sharedCommandCenter()
        self.info_center = MPNowPlayingInfoCenter.defaultCenter()

        self._cmds = [
            (self.cmd_center.togglePlayPauseCommand(), self.on_play_pause),
            (self.cmd_center.playCommand(), self.on_play),
            (self.cmd_center.pauseCommand(), self.on_pause),
            (self.cmd_center.nextTrackCommand(), self.on_next),
            (self.cmd_center.previousTrackCommand(), self.on_previous),
        ]
        for cmd, handler in self._cmds:
            cmd.addTargetWithHandler_(create_handler(cmd, handler))

    def set_playback_property(self, name: PlaybackPropertyName, value: Any):
        if name == PlaybackPropertyName.Metadata:
            value: PlaybackProperties.MetadataBean
            nowplaying_info = self._get_or_create_nowplaying_info()
            nowplaying_info[MPMediaItemPropertyTitle] = value.title
            nowplaying_info[MPMediaItemPropertyArtist] = ', '.join(value.artist)
            nowplaying_info[MPMediaItemPropertyAlbumTitle] = value.album
            self.info_center.setNowPlayingInfo_(nowplaying_info)
        elif name == PlaybackPropertyName.PlaybackStatus:
            value: PlaybackProperties.PlaybackStatus
            # Need to set the rate to 0 when it is paused or stopped,
            # otherwise, macos's position will be incorrect after the player resumes.
            if value in (PlaybackStatus.Paused, PlaybackStatus.Stopped):
                rate = 0
            else:
                rate = self.get_playback_property(PlaybackPropertyName.Rate)
            # Must update the position to keep it accurate. Note that updating the
            # nowplaying info re-paint the UI.
            # HELP: Is there a better way to do this?
            nowplaying_info = self._get_or_create_nowplaying_info()
            nowplaying_info[MPNowPlayingInfoPropertyElapsedPlaybackTime] = \
                self.get_playback_property(PlaybackPropertyName.Position) / 1000
            nowplaying_info[MPNowPlayingInfoPropertyPlaybackRate] = rate
            self.info_center.setNowPlayingInfo_(nowplaying_info)
            self.info_center.setPlaybackState_(PlaybackStatusStateMapping[value])
        elif name == PlaybackPropertyName.Position:
            self._update_property(MPNowPlayingInfoPropertyElapsedPlaybackTime, value/1000)
        elif name == PlaybackPropertyName.Rate:
            self._update_property(MPNowPlayingInfoPropertyDefaultPlaybackRate, 1)
            self._update_property(MPNowPlayingInfoPropertyPlaybackRate, value)
        elif name == PlaybackPropertyName.Duration:
            self._update_property(MPMediaItemPropertyPlaybackDuration, value/1000)
        else:
            # TODO: handle more properties changes
            pass

    async def start(self):
        pass

    def _get_or_create_nowplaying_info(self):
        current = self.info_center.nowPlayingInfo()
        if current is not None:
            nowplaying_info = current.mutableCopy()
        else:
            nowplaying_info = NSMutableDictionary.dictionary()
        return nowplaying_info

    def _update_property(self, name, value):
        nowplaying_info = self._get_or_create_nowplaying_info()
        nowplaying_info[name] = value
        self.info_center.setNowPlayingInfo_(nowplaying_info)



import asyncio
import threading
from typing import Any
from datetime import timedelta

from winsdk.windows.foundation import Uri
from winsdk.windows.foundation.collections import IVector
from winsdk.windows.media import SystemMediaTransportControlsTimelineProperties, SystemMediaTransportControls, \
    SystemMediaTransportControlsDisplayUpdater, MediaPlaybackStatus, MediaPlaybackType, MediaPlaybackAutoRepeatMode, \
    AutoRepeatModeChangeRequestedEventArgs, SystemMediaTransportControlsButtonPressedEventArgs, \
    SystemMediaTransportControlsButton, PlaybackPositionChangeRequestedEventArgs, PlaybackRateChangeRequestedEventArgs, \
    SystemMediaTransportControlsPropertyChangedEventArgs, SystemMediaTransportControlsProperty, \
    ShuffleEnabledChangeRequestedEventArgs
from winsdk.windows.media.playback import MediaPlayer
from winsdk.windows.storage.streams import RandomAccessStreamReference

from aionowplaying import BaseInterface, PropertyName, PlaybackPropertyName
from aionowplaying.interface.base import TrackListPropertyName, PlaybackStatus, PlaybackProperties, LoopStatus, \
    MediaType


def TimeSpan(x):
    return timedelta(microseconds=x/10000)


class WindowsInterface(BaseInterface):
    def __init__(self, name):
        super(WindowsInterface, self).__init__(name)
        self._loop = asyncio.get_event_loop()
        self._running = True
        self._playback_properties = PlaybackProperties()
        self._player = MediaPlayer()
        self._controls: SystemMediaTransportControls = self._player.system_media_transport_controls
        self._updater: SystemMediaTransportControlsDisplayUpdater = self._controls.display_updater
        self._timeline = SystemMediaTransportControlsTimelineProperties()
        self._controls.add_auto_repeat_mode_change_requested(self.auto_repeat_mode_change_requested)
        self._controls.add_button_pressed(self.button_pressed)
        self._controls.add_playback_position_change_requested(self.playback_position_change_requested)
        self._controls.add_playback_rate_change_requested(self.playback_rate_change_requested)
        self._controls.add_property_changed(self.property_changed)
        self._controls.add_shuffle_enabled_change_requested(self.shuffle_change_requested)

    def shuffle_change_requested(self, _, args: ShuffleEnabledChangeRequestedEventArgs):
        shuffle_enabled: bool = args.requested_shuffle_enabled
        if asyncio.iscoroutinefunction(self.on_shuffle):
            self._run_task(self.on_shuffle(shuffle_enabled))
        else:
            self.on_shuffle(shuffle_enabled)

    def property_changed(self, _, args: SystemMediaTransportControlsPropertyChangedEventArgs):
        property_: SystemMediaTransportControlsProperty = args.property
        if property_ == SystemMediaTransportControlsProperty.SOUND_LEVEL:
            if asyncio.iscoroutinefunction(self.on_volume):
                self._run_task(self.on_volume(self._controls.sound_level))
            else:
                self.on_volume(self._controls.sound_level)

    def playback_rate_change_requested(self, _, args: PlaybackRateChangeRequestedEventArgs):
        rate: float = args.requested_playback_rate
        if asyncio.iscoroutinefunction(self.on_rate):
            self._run_task(self.on_rate(rate))
        else:
            self.on_rate(rate)

    def playback_position_change_requested(self, _, args: PlaybackPositionChangeRequestedEventArgs):
        position = args.requested_playback_position
        if self._playback_properties.CanSeek:
            if asyncio.iscoroutinefunction(self.on_set_position):
                self._run_task(self.on_set_position(self._playback_properties.Metadata.id_, int(position.duration / 10000)))
                self._run_task(self.on_seek(int(position.duration / 10000)))
            else:
                self.on_set_position(self._playback_properties.Metadata.id_, int(position.duration / 10000))
                self.on_seek(int(position.duration / 10000))

    def button_pressed(self, _, args: SystemMediaTransportControlsButtonPressedEventArgs):
        button: SystemMediaTransportControlsButton = args.button
        if button == SystemMediaTransportControlsButton.PLAY and self._playback_properties.CanPlay:
            if asyncio.iscoroutinefunction(self.on_play):
                self._run_task(self.on_play())
            else:
                self.on_play()
            self._controls.playback_status = MediaPlaybackStatus.PLAYING
            self._playback_properties.PlaybackStatus = PlaybackStatus.Playing
        if button == SystemMediaTransportControlsButton.PAUSE and self._playback_properties.CanPause:
            if asyncio.iscoroutinefunction(self.on_pause):
                self._run_task(self.on_pause())
            else:
                self.on_pause()
            self._controls.playback_status = MediaPlaybackStatus.PAUSED
            self._playback_properties.PlaybackStatus = PlaybackStatus.Paused
        if button == SystemMediaTransportControlsButton.NEXT and self._playback_properties.CanGoNext:
            if asyncio.iscoroutinefunction(self.on_next):
                self._run_task(self.on_next())
            else:
                self.on_next()
        if button == SystemMediaTransportControlsButton.PREVIOUS and self._playback_properties.CanGoPrevious:
            if asyncio.iscoroutinefunction(self.on_previous):
                self._run_task(self.on_previous())
            else:
                self.on_previous()

    def auto_repeat_mode_change_requested(self, _, args: AutoRepeatModeChangeRequestedEventArgs):
        value = LoopStatus.None_
        mode: MediaPlaybackAutoRepeatMode = args.requested_auto_repeat_mode
        if mode == MediaPlaybackAutoRepeatMode.LIST:
            value = LoopStatus.Playlist
        elif mode == MediaPlaybackAutoRepeatMode.TRACK:
            value = LoopStatus.Track
        if asyncio.iscoroutinefunction(self.on_loop_status):
            self._run_task(self.on_loop_status(value))
        else:
            self.on_loop_status(value)
        self._playback_properties.LoopStatus = value

    def set_property(self, name: PropertyName, value: Any):
        pass

    def set_playback_property(self, name: PlaybackPropertyName, value: Any):
        if name == PlaybackPropertyName.CanPlay:
            self._controls.is_play_enabled = value
            self._playback_properties.CanPlay = value
        elif name == PlaybackPropertyName.CanPause:
            self._controls.is_pause_enabled = value
            self._playback_properties.CanPause = value
        elif name == PlaybackPropertyName.CanGoNext:
            self._controls.is_next_enabled = value
            self._playback_properties.CanGoNext = value
        elif name == PlaybackPropertyName.CanGoPrevious:
            self._controls.is_previous_enabled = value
            self._playback_properties.CanGoPrevious = value
        elif name == PlaybackPropertyName.PlaybackStatus:
            if value == PlaybackStatus.Playing:
                self._controls.playback_status = MediaPlaybackStatus.PLAYING
            elif value == PlaybackStatus.Paused:
                self._controls.playback_status = MediaPlaybackStatus.PAUSED
            elif value == PlaybackStatus.Stopped:
                self._controls.playback_status = MediaPlaybackStatus.STOPPED
            self._playback_properties.PlaybackStatus = value
        elif name == PlaybackPropertyName.Metadata:
            self._update_metadata(value)
            self._playback_properties.Metadata = value
        elif name == PlaybackPropertyName.Shuffle:
            self._controls.shuffle_enabled = value
            self._playback_properties.Shuffle = value
        elif name == PlaybackPropertyName.Rate:
            self._controls.playback_rate = value
            self._playback_properties.Rate = value
        elif name == PlaybackPropertyName.LoopStatus:
            if value == LoopStatus.None_:
                self._controls.auto_repeat_mode = MediaPlaybackAutoRepeatMode.NONE
            elif value == LoopStatus.Playlist:
                self._controls.auto_repeat_mode = MediaPlaybackAutoRepeatMode.LIST
            elif value == LoopStatus.Track:
                self._controls.auto_repeat_mode = MediaPlaybackAutoRepeatMode.TRACK
            self._playback_properties.LoopStatus = value
        elif name == PlaybackPropertyName.Position:
            self._timeline.position = TimeSpan(value * 10000)
            self._controls.update_timeline_properties(self._timeline)

    def _update_metadata(self, value: PlaybackProperties.MetadataBean):
        # update media info
        if value.media_type == MediaType.Image:
            self._updater.type = MediaPlaybackType.IMAGE
        elif value.media_type == MediaType.Video:
            self._updater.type = MediaPlaybackType.VIDEO
        else:
            self._updater.type = MediaPlaybackType.MUSIC
        self._updater.app_media_id = value.id_
        self._updater.music_properties.artist = ','.join(value.artist)
        self._updater.music_properties.title = value.title
        self._updater.music_properties.album_title = value.album
        self._updater.music_properties.genres: IVector
        self._updater.music_properties.genres.replace_all(value.genre)
        if value.cover:  # not None and not empty
            self._updater.thumbnail = RandomAccessStreamReference.create_from_uri(Uri(value.cover))
        self._updater.update()
        # update timeline
        self._timeline.start_time = TimeSpan(0)
        self._timeline.end_time = TimeSpan(value.duration * 10000)
        self._timeline.min_seek_time = TimeSpan(0)
        self._timeline.max_seek_time = TimeSpan(value.duration * 10000)
        self._controls.update_timeline_properties(self._timeline)

    def set_tracklist_property(self, name: TrackListPropertyName, value: Any):
        pass

    def get_playback_property(self, name: PlaybackPropertyName) -> Any:
        return getattr(self._playback_properties, name.value)

    async def start(self):
        # Don't need a background server for Windows
        while self._running:
            await asyncio.sleep(1)
        try:
            self._player.close()
        except:
            pass

    async def stop(self):
        self._running = False

    def _run_task(self, task):
        # Windows callbacks may be invoked in non-main thread, besides,
        # they may run in different threads.
        if threading.current_thread() is not threading.main_thread():
            asyncio.run_coroutine_threadsafe(task, self._loop)
        else:
            asyncio.create_task(task)
